/* $USAGI: ipv6only_test.c,v 1.5 2002/08/06 07:00:37 yoshfuji Exp $ */

/*
 * Copyright (C) 2000 USAGI/WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <errno.h>

/* defines */
#define DEFAULT_SOCKTYPE	SOCK_DGRAM
#define DEFAULT_HOST		"127.0.0.1"
#define DEFAULT_PORT		"12345"

/* types */
enum mode{
	client_mode,
	server_mode
};

#if (!defined __GNUC__ || __GNUC__ < 2 || \
     defined __cplusplus || \
     (__GNUC__ == 2 && __GNUC_MINOR__ < 7))
#define __CONST_SOCKADDR_ARG	const struct sockaddr *
#endif

/* functions */
char *aitop(const struct addrinfo *ai);
int getsock(const struct addrinfo *ai);
struct addrinfo *getai(const char *host, const char *port,
		       sa_family_t family, int socktype, int passive);
int setipv6only(int sk, int on);
void usage(void);

/* variables */
const char *myname;

int main(int argc, char *argv[]){
	int socktype = DEFAULT_SOCKTYPE;
	const char *local4 = NULL, *local6 = NULL;
	const char *remote4 = NULL, *remote6 = NULL;
	const char *lport = DEFAULT_PORT, *rport = DEFAULT_PORT;
	const char *operation = NULL;
	enum mode mode = server_mode;

	struct addrinfo *lai4 = NULL, *lai6 = NULL;
	struct addrinfo *rai4 = NULL, *rai6 = NULL;
	int c;
	const char *p;
	int s4, s6;

	myname = argv[0];

	while((c = getopt(argc, argv, "dl:L:p:r:R:s")) != EOF){
		switch(c){
		case 'd':
			socktype = SOCK_DGRAM;
			break;
		case 'l':
			local4 = optarg;
			break;
		case 'L':
			local6 = optarg;
			break;
		case 'p':
			lport = optarg;
			break;
		case 'P':
			rport = optarg;
			break;
		case 'r':
			remote4 = optarg;
			break;
		case 'R':
			remote6 = optarg;
			break;
		case 's':
			socktype = SOCK_STREAM;
			break;
		default:
			usage();
		}
	}

	argc -= optind;
	argv += optind;

	if (argc != 1)
		usage();
	operation = argv[0];

	if (local4 || local6){
		if (!local4) local4 = local6;
		if (!local6) local6 = local4;
	}
	if (remote4 || remote6){
		if (!remote4) remote4 = remote6;
		if (!remote6) remote6 = remote4;
	}

	for (p = operation; *p; p++){
		switch(*p){
		case 's':
		case 'u':
			break;
		case 'l':
		case 'L':
			break;
		case '4':
		case '6':
			mode = client_mode;
			break;
		case 'x':
			break;
		default:
			usage();
		}
	}

	lai4 = getai(local4 , lport, PF_INET , socktype, 1);
	lai6 = getai(local6 , lport, PF_INET6, socktype, 1);
	rai4 = getai(remote4, rport, PF_INET , socktype, 0);
	rai6 = getai(remote6, rport, PF_INET6, socktype, 0);

	printf("# mode        : %s\n", mode == client_mode ? "client" : "server");
	printf("# socktype    : %s\n", socktype == SOCK_STREAM ? "stream" : "datagram");
	printf("# local addr. : ipv4 = %s, ", aitop(lai4));
	printf(                "ipv6 = %s\n", aitop(lai6));
	if (mode == client_mode){
		printf("# remote addr.: ipv4 = %s, ", aitop(rai4));
		printf(                "ipv6 = %s\n", aitop(rai6));
	}

	s4 = getsock(lai4);
	printf("socket for ipv4 : %d\n", s4);
	s6 = getsock(lai6);
	printf("socket for ipv6 : %d\n", s6);

	for (p = operation; *p; p++) {
		switch(*p){
		case 's':
		case 'u':
		    {
#ifdef IPV6_V6ONLY
			int on = (*p == 's') ? 1 : 0;
			printf("%s IPV6_V6ONLY socket option for ipv6 socket: ",
				on ? "set" : "unset");
			if (setsockopt(s6, IPPROTO_IPV6, 
					IPV6_V6ONLY, &on, sizeof(on)) < 0) {
				printf("%s\n", strerror(errno));
			} else {
				printf("ok\n");
			}
#else
			printf("IPV6_V6ONLY socket options is not supported.\n");
#endif
			break;
		    }
		case 'l':
		case 'L':
		case '4':
		case '6':
		    {
			int sk;
			char *ip46, *bc;
			struct addrinfo *lai;
			int (*call)(int sk, __CONST_SOCKADDR_ARG sa, socklen_t salen);
			if (*p == 'L' || *p == '6') {
				sk = s6;
				ip46 = "ipv6";
				lai = lai6;
			} else {
				sk = s4;
				ip46 = "ipv4";
				lai = lai4;
			}
			if (*p == '4' || *p == '6') {
				bc = "connect";
				call = connect;
			} else {
				bc = "bind";
				call = bind;
			}
			printf("%s %s socket: ", bc, ip46);
			if (call(sk, lai->ai_addr, lai->ai_addrlen) < 0) {
				printf("%s\n", strerror(errno));
			} else {
				printf("ok\n");
			}
			break;
		    }
		case 'x':
		    	sleep(10);
			break;
		}
	}
}

char *aitop(const struct addrinfo *ai){
	char abuf[NI_MAXHOST], pbuf[NI_MAXSERV];
	static char buf[NI_MAXHOST+6+NI_MAXSERV];
	if (getnameinfo(ai->ai_addr, ai->ai_addrlen,
			abuf, sizeof(abuf), pbuf, sizeof(pbuf),
			NI_NUMERICHOST|NI_NUMERICSERV)){
		strcpy(abuf, "???");
		strcpy(pbuf, "???");
	}
	sprintf(buf, "%s port %s", abuf, pbuf);
	return buf;
}
int getsock(const struct addrinfo *ai){
	int s;
	s = socket(ai->ai_family, ai->ai_socktype, ai->ai_protocol);
	if (s < 0){
		perror("socket");
		exit(1);
	}
	return s;
}
struct addrinfo *getai(const char *host, const char *port,
		       sa_family_t family, int socktype, int passive){
	struct addrinfo hints, *ai;
	int gai;
	memset(&hints, 0, sizeof(hints));
	hints.ai_family = family;
	hints.ai_socktype = socktype;
	hints.ai_flags = passive ? AI_PASSIVE : 0;
	gai = getaddrinfo(host, port, &hints, &ai);
	if (gai){
		fprintf(stderr,
			"getai(): cannot resolve host '%s' port '%s': %s\n",
			host, port, gai_strerror(gai));
		exit(1);
	}
	return ai;
}

void usage(void){
	fprintf(stderr,
		"Usage: %s [options] [operation...]\n"
		"\n"
		"Options:\n"
		"\t"	"-s          use stream socket\n"
		"\t"	"-d          use datagram socket\n"
		"\t"	"-p port     local port to bind\n"
		"\t"	"-P port     remote port to connect\n"
		"\t"	"-l v4addr   local ipv4 address to bind\n"
		"\t"	"-L v6addr   local ipv6 address to bind\n"
		"\t"	"-r v4addr   remote ipv4 address to connect\n"
		"\t"	"-R v6addr   remote ipv6 address to connect\n"
		"\n"
		"Operation:\n"
		"\t"	"s           set IPV6_ONLY sockopt\n"
		"\t"	"u           unset IPV6_ONLY sockopt\n"
		"\t"	"l           bind ipv4\n"
		"\t"	"L           bind ipv6\n"
		"\t"	"4           connect remote ipv4\n"
		"\t"	"6           connect remote ipv6\n"
		"\n",
		myname
	);
	exit(1);
}
